from glasskit import ctx


class BaseProvider:

    PROVIDER_NAME = None
    PROVIDER_MAP = {}
    state = None

    def __init__(self):
        from .config import get_conf

        if self.PROVIDER_NAME is None:
            raise RuntimeError("can't instantiate BaseProvider abstract class")
        cfg = get_conf(self.PROVIDER_NAME)
        if cfg is None:
            raise RuntimeError(
                f"no configuration for provider {self.PROVIDER_NAME} given"
            )
        self.client_id = cfg[0]
        self.client_secret = cfg[1]
        self.redirect_uri = cfg[2]
        self.extra = cfg[3]

    def authorize_uri(self, state):
        raise NotImplementedError("authorize_uri method must be overriden")

    def acquire_token(self, code):
        raise NotImplementedError("acquire_token method must be overriden")

    def get_user_data(self, token):
        raise NotImplementedError("_fetch_user_data method must be overriden")

    def register(self):
        if self.PROVIDER_NAME is None:
            raise RuntimeError(
                f"error registering {self.__class__.__name__} provider: PROVIDER_NAME is not defined"
            )
        BaseProvider.PROVIDER_MAP[self.PROVIDER_NAME] = self
        ctx.log.info(f'OAuth2 provider "{self.PROVIDER_NAME}" registered')

    @classmethod
    def get_provider(cls, provider_name):
        if provider_name not in cls.PROVIDER_MAP:
            raise RuntimeError(f"provider {provider_name} not found or not registered")
        return cls.PROVIDER_MAP[provider_name]

    def provider_info(self):
        return {
            "provider_name": self.PROVIDER_NAME,
            "authorize_uri": self.authorize_uri(state=self.PROVIDER_NAME),
            **self.extra,
        }

    @classmethod
    def list_provider_info(cls):
        return [p.provider_info() for p in cls.PROVIDER_MAP.values()]



from glasskit import ctx, json
from glasskit.errors import OAuthError
from .provider import BaseProvider


class FacebookProvider(BaseProvider):

    PROVIDER_NAME = "facebook"

    def authorize_uri(self, state):
        if type(state) != str:
            state = json.dumps(state)
        return (
            f"https://www.facebook.com/v7.0/dialog/oauth?client_id={self.client_id}"
            f"&redirect_uri={self.redirect_uri}&state={state}"
        )

    def acquire_token(self, code):
        uri = (
            f"https://graph.facebook.com/v7.0/oauth/access_token?"
            f"client_id={self.client_id}&client_secret={self.client_secret}"
            f"&redirect_uri={self.redirect_uri}&code={code}"
        )
        ctx.log.debug("about to perform token request via %s", uri)
        resp = requests.get(uri)
        if resp.status_code != 200:
            raise OAuthError(
                "non-successful status code from oauth provider access_token handler",
                payload={"status_code": resp.status_code},
            )
        data = resp.json()
        if "access_token" not in data:
            raise OAuthError("access_token is not in response data")
        return data["access_token"]

    def get_user_data(self, token):
        uri = f"https://graph.facebook.com/v7.0/me?fields=first_name,last_name,id&access_token={token}"
        resp = requests.get(uri)
        if resp.status_code != 200:
            raise OAuthError(
                "non-successful status code from oauth provider userdata handler",
                payload={"status_code": resp.status_code},
            )
        attrs = resp.json()
        return {
            "ext_id": f"fb_{attrs['id']}",
            "username": f"fb_{attrs['id']}",
            "first_name": attrs["first_name"],
            "last_name": attrs["last_name"],
        }



from . import Command


class Run(Command):
    def init_argument_parser(self, parser):
        parser.add_argument("-H", "--host", default="127.0.0.1")
        parser.add_argument("-P", "--port", type=int, default=5000)

    def run(self):
        self.app.run(debug=True, host=self.args.host, port=self.args.port)



import requests
from urllib.parse import parse_qs
from glasskit import ctx, json
from glasskit.errors import OAuthError
from .provider import BaseProvider


class GithubProvider(BaseProvider):

    PROVIDER_NAME = "github"

    def authorize_uri(self, state):
        if type(state) != str:
            state = json.dumps(state)
        self.state = state
        return (
            f"https://github.com/login/oauth/authorize?client_id={self.client_id}"
            f"&redirect_uri={self.redirect_uri}&state={state}"
        )

    def acquire_token(self, code):
        uri = f"https://github.com/login/oauth/access_token"
        payload = {
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "code": code,
            "redirect_uri": self.redirect_uri,
            "state": self.state,
        }
        ctx.log.debug("about to perform token request via %s", uri)
        resp = requests.post(uri, json=payload)
        if resp.status_code != 200:
            raise OAuthError(
                "non-successful status code from oauth provider access_token handler",
                payload={"status_code": resp.status_code},
            )
        try:
            data = parse_qs(resp.content)
        except Exception as e:
            raise OAuthError(f"error parsing access token data from github {e}")
        if b"access_token" not in data:
            raise OAuthError("access_token is not in response data")
        return data[b"access_token"][0].decode()

    def get_user_data(self, token):
        uri = "https://api.github.com/user"
        headers = {"Authorization": f"token {token}"}
        resp = requests.get(uri, headers=headers)
        data = resp.json()

        first_name = None
        last_name = None
        name = data["name"].split()

        if name:
            first_name = name[0]
            last_name = " ".join(name[1:])

        return {
            "ext_id": f"gh_{data['id']}",
            "username": data["login"],
            "avatar_url": data["avatar_url"],
            "first_name": first_name,
            "last_name": last_name,
            "email": data["email"],
        }



import requests
from glasskit import ctx, json
from glasskit.errors import OAuthError
from .provider import BaseProvider


class YandexProvider(BaseProvider):

    PROVIDER_NAME = "yandex"

    def authorize_uri(self, state):
        if type(state) != str:
            state = json.dumps(state)
        return f"https://oauth.yandex.ru/authorize?response_type=code&client_id={self.client_id}&redirect_uri={self.redirect_uri}&state={state}"

    def acquire_token(self, code):
        uri = "https://oauth.yandex.ru/token"
        payload = {
            "code": code,
            "grant_type": "authorization_code",
            "client_id": self.client_id,
            "client_secret": self.client_secret,
        }
        ctx.log.debug("about to perform token request via %s", uri)
        resp = requests.post(uri, payload)
        if resp.status_code != 200:
            raise OAuthError(
                "non-successful status code from oauth provider access_token handler",
                payload={"status_code": resp.status_code},
            )
        data = resp.json()

        if "access_token" not in data:
            raise OAuthError("access_token is not in response data")
        return data["access_token"]

    def get_user_data(self, token):
        uri = "https://login.yandex.ru/info?format=json"
        headers = {"Authorization": f"OAuth {token}"}
        resp = requests.get(uri, headers=headers)
        if resp.status_code != 200:
            raise OAuthError(
                "non-successful status code from oauth provider userdata handler",
                payload={"status_code": resp.status_code},
            )
        attrs = resp.json()

        user_data = {
            "ext_id": f"yandex_{attrs['id']}",
            "username": attrs["display_name"],
            "first_name": attrs["first_name"],
            "last_name": attrs["last_name"],
        }

        if not attrs["is_avatar_empty"]:
            user_data[
                "avatar_url"
            ] = f"//avatars.mds.yandex.net/get-yapic/{attrs['default_avatar_id']}/islands-200"

        return user_data



import functools
import re

from collections import namedtuple
from math import ceil
from flask import request, make_response
from datetime import datetime

from . import ctx, json

ArithmeticExpression = namedtuple("_ArithmeticExpression", field_names=["op", "value"])

DEFAULT_DOCUMENTS_PER_PAGE = 10
ARITHMETIC_OPS = (
    "eq",
    "lt",
    "gt",
    "lte",
    "gte",
    "ne",
)
INTEGER_EXPRESSION = re.compile(r"^\d+$")


def get_page(nopaging=False):
    if nopaging:
        return None
    if "_page" in request.values:
        page = request.values["_page"]
    else:
        page = 1
    try:
        page = int(page)
    except (TypeError, ValueError):
        page = 1
    return page


def get_limit(nopaging=False):
    if nopaging:
        return None
    default_limit = ctx.cfg.get("documents_per_page", DEFAULT_DOCUMENTS_PER_PAGE)
    if "_limit" in request.values:
        limit = request.values["_limit"]
        try:
            limit = int(limit)
        except ValueError:
            limit = default_limit
    else:
        limit = default_limit
    return limit


def get_request_fields(default_fields=None):
    if "_fields" in request.values:
        return request.values["_fields"].split(",")
    return default_fields


def boolean(value):
    return str(value).lower() in ("yes", "true", "1")


def get_boolean_request_param(param_name):
    return boolean(request.values.get(param_name))


def paginated(data, page=None, limit=None, extra=None, transform=None):

    nopaging = get_boolean_request_param("_nopaging")
    if page is None:
        page = get_page(nopaging)
    if limit is None:
        limit = get_limit()

    if isinstance(data, list):
        count = len(data)
        if limit is not None and page is not None:
            data = data[(page - 1) * limit : page * limit]
    elif hasattr(data, "count"):
        if limit is not None and page is not None:
            data = data.skip((page - 1) * limit).limit(limit)
        #
        # count() must be called after skip and limit are set:
        #
        # This doesn't change the result of count() however,
        # in some cases (i.e. not pymongo cursor but a cursor-like
        # elasticsearch result object), calling count() leads to
        # an actual query while setting skip and limit is supposed
        # to take place before the actual search call
        count = data.count()
    else:
        raise RuntimeError("paginated() accepts either cursor objects or lists")

    if transform is not None:
        data = [transform(x) for x in data]

    total_pages = ceil(count / limit) if limit is not None else None

    result = {"page": page, "total_pages": total_pages, "count": count, "data": data}

    if extra is not None and hasattr(extra, "items"):
        for k, v in extra.items():
            if k not in result:
                result[k] = v

    return result


def default_transform(fields=None):
    def transform(x):
        return x.to_dict(fields=get_request_fields(fields))

    return transform


def get_int_request_param(param_name):
    value = request.values.get(param_name)
    if value is None:
        return None
    try:
        value = int(value)
    except ValueError:
        return None
    return value


def get_utcts_request_param(param_name):
    ts = get_int_request_param(param_name)
    if not ts:
        return None
    return datetime.utcfromtimestamp(ts)


def filter_expr(flt, regexp=True):
    if flt.find(" ") >= 0:
        tokens = flt.split()
        return {"$in": tokens}
    if regexp:
        try:
            return re.compile(flt, re.IGNORECASE)
        # re.compile's can throw multiple different exceptions. We do not care what went wrong
        except Exception:  # pylint: disable=broad-except
            return ""
    else:
        return flt


def arithmetic_expr(expr, try_int=True):
    tokens = expr.split(":")
    expr_op = "eq"
    if len(tokens) == 2 and tokens[0] in ARITHMETIC_OPS:
        expr_op = tokens[0]
        expr = tokens[1]

    if try_int and INTEGER_EXPRESSION.match(expr):
        expr = int(expr)

    return ArithmeticExpression(op=expr_op, value=expr)


def json_body_required(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        from .errors import InputDataError

        if request.json is None:
            raise InputDataError("json data is missing")
        return func(*args, **kwargs)

    return wrapper


def json_response(data, code=200):
    json_kwargs = {}
    if ctx.cfg.get("debug"):
        json_kwargs["indent"] = 4
    return make_response(
        json.dumps(data, **json_kwargs), code, {"Content-Type": "application/json"}
    )



import os
from copy import copy
from typing import Optional
from glasskit import ctx
from glasskit.errors import ConfigurationError
from .provider import BaseProvider


def get_conf(provider_name):

    oacfg = ctx.cfg.get("oauth")
    if not oacfg:
        raise ConfigurationError("oauth is not configured")

    gcfg = oacfg.get("_global")

    if provider_name not in oacfg:
        return None

    provider_cfg = {**gcfg, **copy(oacfg[provider_name])}
    if "client_id" not in provider_cfg:
        raise ConfigurationError(f"oauth provider {provider_name} id field missing")
    if "client_secret" not in provider_cfg:
        raise ConfigurationError(f"oauth provider {provider_name} secret field missing")
    if "provider_name" not in provider_cfg:
        provider_cfg["provider_name"] = provider_name
    if "redirect_uri" not in provider_cfg:
        raise ConfigurationError(
            f"oauth provider {provider_name} redirect_uri field missing and no"
            f" global redirect_uri found"
        )

    client_id = provider_cfg["client_id"]
    client_secret = provider_cfg["client_secret"]
    redirect_uri = provider_cfg["redirect_uri"]
    del provider_cfg["client_id"]
    del provider_cfg["client_secret"]
    del provider_cfg["redirect_uri"]

    return client_id, client_secret, redirect_uri, provider_cfg


def setup_providers(relative_path: Optional[str] = None) -> None:
    from . import FacebookProvider, YandexProvider, GithubProvider

    providers = [FacebookProvider, YandexProvider, GithubProvider]

    if relative_path:
        module_prefix = ".".join(relative_path.split("/")).strip(".")

        for filename in os.listdir(relative_path):
            if filename.endswith(".py") and not filename.startswith("__"):
                module_name = filename[:-3]
                module_name = f"{module_prefix}.{module_name}"
                module = importlib.import_module(module_name)
                for obj_name in dir(module):
                    obj = getattr(module, obj_name)
                    if (
                        inspect.isclass(obj)
                        and issubclass(obj, BaseProvider)
                        and obj != BaseProvider
                    ):
                        providers.append(obj)

    for provider in providers:
        if get_conf(provider.PROVIDER_NAME):
            provider().register()
        else:
            ctx.log.info(
                "Skipping OAuth2 provider %s, no configuration found",
                provider.PROVIDER_NAME,
            )
